import ctypes
from ctypes import cdll
import os.path
from .token import Token
import json
from datetime import datetime
from .nextfloat import next_up
import sys

class NLexWrappedObject(object):
    class ValueStruct(ctypes.Structure):
        _fields_ = [
            ("start", ctypes.POINTER(ctypes.c_char)),
            ("length", ctypes.c_int),
            ("tag", ctypes.c_char_p),
            ("errc", ctypes.c_char),
            ("metadata", ctypes.c_ubyte),
            ("pos", ctypes.POINTER(ctypes.c_char)),
            ("allocd", ctypes.c_int),
        ]
        def __repr__(self):
            return f"NLexWrappedObject.ValueStruct(start={self.start}, length={self.length}, tag={self.tag}, errc={self.errc}, metadata={self.metadata})"

    def __init__(self, path=os.path.join(os.path.dirname(os.path.abspath(__file__)), 'libtokenise.so'), log=False):
        self.log = log
        self.__lib = cdll.LoadLibrary(path)
        self._fed = None
        self._m_value = NLexWrappedObject.ValueStruct()
        self._nlex_feed = getattr(self.__lib, '__nlex_feed')
        self._nlex_feed.argtypes = (ctypes.c_char_p,)
        self._nlex_root = getattr(self.__lib, '__nlex_root')
        self._nlex_root.argtypes = (ctypes.POINTER(NLexWrappedObject.ValueStruct),)
        self._nlex_distance = getattr(self.__lib, '__nlex_distance')
        self.__nlex_skip = getattr(self.__lib, '__nlex_skip')
        self.__has_postag = ctypes.c_int.in_dll(self.__lib, '__nlex_has_tagpos')
        self.can_split_sentences = self.__has_postag
        self.__postag_gram = None
        if self.__has_postag:
            self.__postag_gram = ctypes.c_int.in_dll(self.__lib, '__nlex_tagpos_gram')
            self.__sentence_delimiter = ctypes.string_at(ctypes.addressof(ctypes.c_char.in_dll(self.__lib, '__nlex_ptag')))
            # self.token_array_space = (NLexWrappedObject.ValueStruct * 4096)() # 4096 tokens in a sentence...?
            self._create_postagger()

        self.__has_normaliser = True
        self.__last_offset = -1
        self.total = 0
        try:
            self._nlex_pure_normalise = getattr(self.__lib, '__nlex_pure_normalise')
            self._nlex_pure_normalise.restype = ctypes.c_char
        except AttributeError:
            self.__has_normaliser = False

    def _create_postagger(self):
        def next_sentence(cleanup):
            if not self._fed:
                raise Exception("NLexWrappedObject.__next_tagged_sentence called before __feed")
            sentence = []
            token = self.__next_token(cleanup)
            sentence.append(token)
            while not token or token.tag != self.__sentence_delimiter:
                token = self.__next_token(cleanup)
                sentence.append(token)
            return sentence

        def sentences(clean=True):
            i = 0
            while True:
                try:
                    x = self.next_sentence(clean)
                except Exception as e:
                    return
                if not x:
                    return
                yield x
                i += 1

        # TODO: move this out
        def ptag_sentences(clean=True):
            raise NotImplemented

        self.next_sentence = next_sentence
        self.sentences = sentences
        self.pos_tagged_sentences = ptag_sentences

        # init pos tagger
        # self.__nlex_load_postagger = getattr(self.__lib, '__nlex_load_tagpos')
        # self.__nlex_unload_postagger = getattr(self.__lib, '__nlex_unload_tagpos')
        #self.__nlex_load_postagger()

    def __del__(self):
        # if self.__has_postag:
            # self.__nlex_unload_postagger()
        pass

    def __feed(self, string):
        self._fed = ctypes.create_string_buffer(bytes(string, 'utf-8') if isinstance(string, str) else string)
        self._fed_ptr = ctypes.cast(self._fed, ctypes.POINTER(ctypes.c_char))
        self._nlex_feed(self._fed)

    def feed(self, string):
        self.__feed(string)
        self.fedlen = len(string.encode('utf-8'))

    def __next_token(self, cleanup):
        if not self._fed:
            raise Exception("NLexWrappedObject.__next_token called before __feed")

        self._nlex_root(ctypes.pointer(self._m_value))
        offset = self._nlex_distance()-self._m_value.length
        if cleanup:
            if self._m_value.errc != b'\0': # TODO: can_produce_token()
                self._fed = None
                return None

        if offset == self.__last_offset or offset >= self.fedlen:
            if offset >= self.fedlen:
                self._fed = None
                return None
            if self.log:
                print('early terminate at', offset, 'out of', self.fedlen, 'on', self._fed[offset:offset+4].decode('utf8', 'ignore'), file=sys.stderr)
            self.__nlex_skip()

        self.__last_offset = offset
        if self.log:
            print('offset', offset, 'out of', self.fedlen, file=sys.stderr)
        if self._m_value.length == 0:
            return self.__next_token(cleanup)
        return Token(
            value=(ctypes.c_char * self._m_value.length).from_address(ctypes.addressof(self._m_value.start.contents)),
            length=self._m_value.length,
            tag=self._m_value.tag,
            metadata=self._m_value.metadata,
            offset=offset
        )

    def __next_normalised_char(self):
        if not self.__has_normaliser:
            raise Exception("NLex object not built with `option pure_normaliser on`")
        if not self._fed:
            raise Exception("NLexWrappedObject.__next_normalised_char called before __feed")

        return self._nlex_pure_normalise()

    def next_token(self, clean=True):
        return self.__next_token(clean)

    def next_normalised_char(self):
        return self.__next_normalised_char()

    def tokens(self, clean=True):
        i = 0
        while True:
            # if i > 1000:
            #     return
            x = self.__next_token(clean)
            if x is None:
                return
            yield x
            i += 1

    def normalise_all(self):
        x = b''
        buf = b''
        while x != b'\0':
            buf += x
            x = self.__next_normalised_char()
        self._fed = None
        return buf

    def process_documents(self, ds, is_json=False, to_json=False, filename='-', clean=True):
        s = ds
        if is_json:
            s = json.loads(ds)
        if not isinstance(s, (list, tuple)):
            s = [s]
        docs = []
        res = {
            'description': 'NLex',
            'filename': filename,
            'rundate': datetime.now().strftime('%D %T'),
            'documents': docs,
        }
        for di,req in enumerate(s):
            self.next_id = 0
            if hasattr(req, 'text'):
                self.feed(req['text'])
            else:
                self.feed(req)
            if self.can_split_sentences:
                docs.append({
                    'id': di,
                    'sentences': [{
                        'id': sid,
                        'tokens': list(x.desanitify(self) for x in sentence)
                    } for sid,sentence in enumerate(
                            self.sentences(clean))
                    ]
                })
            else:
                docs.append({
                    'id': di,
                    'sentences': [{
                        'id': 0, # we have no idea :shrug:
                        'tokens': list(x.desanitify(self) for x in self.tokens(clean))
                    }]
                })
        if isinstance(to_json, str):
            with open(to_json, 'w', encoding='utf-8') as f:
                json.dump(res, f)
            return None

        return json.dumps(res) if to_json else res

    def next_id():
        doc = "next token id (not unique per token)"

        def fget(self):
            x = self._next_id
            self._next_id += 1 # next_up(self._next_id, 10)
            self.total += 1
            return x

        def fset(self, value):
            self._next_id = value

        def fdel(self):
            del self._next_id

        return locals()
    next_id = property(**next_id())



import math
import struct

def next_up(x, i):
    # NaNs and positive infinity map to themselves.
    if math.isnan(x) or (math.isinf(x) and x > 0):
        return x

    # 0.0 and -0.0 both map to the smallest +ve float.
    if x == 0.0:
        x = 0.0

    n = struct.unpack('<q', struct.pack('<d', x))[0]
    if n >= 0:
        n += i
    else:
        n -= i
    return struct.unpack('<d', struct.pack('<q', n))[0]



from wrap.wrapper import NLexWrappedObject
from pprint import pprint
import json
import time
import sys

wo = NLexWrappedObject("../tokenizer")
ns = time.time_ns()
ts = wo.process_json('[]', clean=False)
ns2 = time.time_ns()
print(ts)
dt = ns2 - ns
print("Read", wo.total, "tokens in",
      dt/1000_000, 'ms', 'which is',
      int(wo.total/dt*1000_000_000),
      'tokens/s',
      file=sys.stderr)



from .wrapper import NLexWrappedObject
from textwrap import dedent
import os
import requests
import zipfile
import hashlib
import platform

class NLexTokenizerCreationException(Exception):
    pass

_default_options = {
        'Linux': {'output_file': 'tokenizer.so', 'sys': ''},
        'Windows': {'output_file': 'tokenizer.dll', 'sys': 'windows'},
    }.get(platform.system(), None)

if _default_options is None:
    print(f"Running on unsupported system ({platform.system()}), this is likely to blow up")
    _default_options = {'output_file': 'tokenizer', 'sys': ''}

class callout:
    @staticmethod
    def run(source, arguments, compiler_server):
        class RunResult:
            def __init__(self, diagnostics, identifier, dot_data):
                self.diagnostics = diagnostics
                self.identifier = identifier
                self.dot_data = dot_data

        res = requests.post(compiler_server + '/run', json={'arguments':arguments, 'source':source}).json()
        return RunResult(diagnostics=res['diagnostics'], identifier=res['identifier'], dot_data=res['dot_data'])

    @staticmethod
    def download_url(url, save_path, chunk_size=128):
        r = requests.get(url, stream=True)
        with open(save_path, 'wb') as fd:
            for chunk in r.iter_content(chunk_size=chunk_size):
                fd.write(chunk)

    @staticmethod
    def compile_and_download(identifier, output_file, compiler_server):
        res = requests.post(compiler_server + '/compile', json={'job_id':identifier}).json()
        if not res['ok']:
            raise NLexTokenizerCreationException("Compile error")
        callout.download_url(compiler_server + '/download?file_id=' + res['file_id'], output_file + '.zip')
        if os.path.exists(output_file): os.remove(output_file)
        with zipfile.ZipFile(output_file + '.zip', 'r') as f:
            try:
                try:
                    f.extract('tokenizer.dll')
                    os.rename('tokenizer.dll', output_file)
                except:
                    f.extract('tokenizer.so')
                    os.rename('tokenizer.so', output_file)
                os.chmod(output_file, 0o755)
                return os.path.realpath(output_file)
            except:
                raise



def NLexTokenizer(*args,
        arch='x64', vendor='', sys=_default_options['sys'], target='',
        object_format='', library='on', cpu='generic', relocation_model='pic',
        features='', output_file=_default_options['output_file'],
        compiler_server='https://nlex.herokuapp.com', source_file=None):
    if len(args) == 0:
        return lambda fn: NLexTokenizer(fn,
                arch=arch, vendor=vendor, sys=sys, target=target, relocation_model=relocation_model,
                object_format=object_format, library=library, compiler_server=compiler_server,
                cpu=cpu, features=features, output_file=output_file, source_file=source_file)

    source = None
    fn = args[0]
    if not source_file:
        doc = dedent(fn.__doc__)
        if not doc:
            raise NLexTokenizerCreationException(f"Function {fn.__name__} does not have a docstring")
        source = doc
    else:
        with open(source_file, 'r') as f:
            source = f.read()

    compile_options = {
        'target_vendor': vendor,
        'target_sys': sys,
        'object_format': object_format,
        'library': library,
        'relocation_model': relocation_model,
        'target': target
    }

    path = os.path.realpath(output_file)
    hashpath = os.path.join(os.path.dirname(path), f'.{output_file}.hash')
    hash_data = source + '\n' + repr(compile_options)
    hsh = hashlib.sha512(bytes(hash_data, 'utf8')).hexdigest()
    if os.path.exists(hashpath) and os.path.exists(output_file):
        with open(hashpath, 'r') as f:
            if f.read() == hsh:
                return lambda inp: fn(inp, NLexWrappedObject(path).process_documents)

    res = callout.run(source, compile_options, compiler_server)
    if res.diagnostics:
        raise NLexTokenizerCreationException(f"Error: {res.diagnostics}")

    path = callout.compile_and_download(res.identifier, output_file, compiler_server)

    with open(hashpath, 'w+') as f:
        f.write(hsh)

    return lambda inp: fn(inp, NLexWrappedObject(path).process_documents)




from distutils.core import setup
setup(
  name = 'nlex',
  packages = ['nlex', 'nlex.wrap', 'example'],
  version = '0.0.3.3',
  license='bsd-2-clause',
  description = 'nlex remote compiler wrapper',
  author = 'AnotherTest',
  author_email = 'devanothertest@gmail.com',
  url = 'https://github.com/alimpfard/nlp-lex',
  download_url = 'https://github.com/alimpfard/nlp-lex/archive/pywrapper-0.0.3.tar.gz',
  keywords = [],
  install_requires=[
          'requests',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 3',
  ],
)



class Token(object):
    def __init__(self, value, length, tag, metadata, offset):
        self.raw = value[:length]
        self.length = length
        self.tag = tag[:]
        self.metadata = metadata
        self.offset = offset
        self.pos = None

    def desanitify(self, nlex):
        return {
            'id': nlex.next_id * 10,
            'raw_token': repr(self.raw),
            'token': self.raw.decode('utf-8', 'ignore'),
            'type': self.tag.decode('utf-8', 'ignore'),
            'delete_ref': [], # we don't support this
            'log': 'correct', # not quite sure if it can handle nulls
            'is_stopword': bool(self.metadata&1),
            'POS': self.pos if self.pos else '',
            'stem': ''
        }
    @property
    def token(self):
        return self.raw.decode('utf-8', 'ignore')

    def __repr__(self):
        return f'Token(value={repr(self.raw)}, length={self.length}, tag={self.tag}, metadata={self.metadata}, offset={self.offset})'
